//------------------------------------------------------------------------------------------------+
// Filename: Localization.cpp
// Content : ȈʒusRTR|[lg̊֐`
//------------------------------------------------------------------------------------------------+

#include "Localization.h"
#include <iostream>

using namespace std;


//-------------------------------------------------------------------------------------------------+
// }N̒`
//-------------------------------------------------------------------------------------------------+
#define Iblue(i,x,y) ((uchar*)((i)->imageData + (i)->widthStep*(y)))[3*(x)]		// 24rbgJ[A
#define Igreen(i,x,y) ((uchar*)((i)->imageData + (i)->widthStep*(y)))[3*(x)+1]	// 24rbgJ[A
#define Ired(i,x,y) ((uchar*)((i)->imageData + (i)->widthStep*(y)))[3*(x)+2]	// 24rbgJ[A


//------------------------------------------------------------------------------------------------+
// Module specification
//------------------------------------------------------------------------------------------------+
static const char* localization_spec[] =
{
    "implementation_id", "Localization",
    "type_name",         "Localization",
    "description",       "Localization",
    "version",           "1.0.0",
    "vendor",            "aisl.cs.tut.ac.jp",
    "category",          "Category",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "0",
    "language",          "C++",
    "lang_type",         "compile",
	// Configuration variables
	"conf.default.Particle", "100",
	"conf.default.debug_window", "0",
	"conf.default.USE_INIT_POSE", "0",
	"conf.default.INIT_X", "0.0",
	"conf.default.INIT_Y", "0.0",
	"conf.default.INIT_HEADING", "0.0",
	// Widget
	"conf.__widget__.Particle", "text",
	"conf.__widget__.debug_window", "text",
	"conf.__widget__.USE_INIT_POSE", "text",
	"conf.__widget__.INIT_X", "text",
	"conf.__widget__.INIT_Y", "text",
	"conf.__widget__.INIT_HEADING", "text",
	// Constraints
    ""
};


//-------------------------------------------------------------------------------------------------+
// RXgN^
//-------------------------------------------------------------------------------------------------+
Localization::Localization(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_RangeDataIn("RangeData", m_RangeData),
    m_RobotPoseIn("RobotPose", m_RobotPose),
    m_ImageIn("Image", m_Image),
    m_StereoDataIn("StereoData", m_StereoData),
    m_EstimatedPoseOut("EstimatedPose", m_EstimatedPose),
    m_PosePairOut("PosePair", m_PosePair),
    m_AbsoluteMapServicePort("AbsoluteMapService")
{
	cerr << "--------------------------------------------------\n";
	cerr << "  Localization (Ver. 5th December 2011)\n";
	cerr << "--------------------------------------------------\n\n";

	input = 0;
	kslam = 0;
	cTimer = 0;

	timeBeginPeriod( 1 );		// x1msɐݒ
}


//-------------------------------------------------------------------------------------------------+
// fXgN^
//-------------------------------------------------------------------------------------------------+
Localization::~Localization(){

	delete input;
	delete kslam;
	delete cTimer;
}


//-------------------------------------------------------------------------------------------------+
// vON̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t Localization::onInitialize(){

	// Set InPort buffers
	addInPort("RangeData", m_RangeDataIn);
	addInPort("RobotPose", m_RobotPoseIn);
	addInPort("Image", m_ImageIn);
	addInPort("StereoData", m_StereoDataIn);

	// Set OutPort buffer
	addOutPort("EstimatedPose", m_EstimatedPoseOut);
	addOutPort("PosePair", m_PosePairOut);

	// Set service consumers to Ports
	m_AbsoluteMapServicePort.registerConsumer("AbsoluteMapService", "MRFC::AbsoluteMapService", m_AbsoluteMapService);

	// Set CORBA Service Ports
	addPort(m_AbsoluteMapServicePort);

	// Bind variables and configuration variable
	bindParameter( "Particle", m_Particle_num, "100" );
	bindParameter( "debug_window", m_debug_window, "0" );
	bindParameter( "USE_INIT_POSE", m_USE_INIT_POSE, "0" );
	bindParameter( "INIT_X", m_INIT_X, "0.0" );
	bindParameter( "INIT_Y", m_INIT_Y, "0.0" );
	bindParameter( "INIT_HEADING", m_INIT_HEADING, "0.0" );

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// ANeBx[g̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t Localization::onActivated(RTC::UniqueId ec_id){

	delete input;
	delete kslam;

	input = new SLAMinput;

	while( !InputPose() ){

		cerr << "Waiting input initial pose.\n";
		Sleep( 1000 );
	}

	kslam = new KSLAM( m_Particle_num, input, false, m_debug_window, InitGlobalMap() );

	cTimer = new TIMER::TIMER_CLASS();		// ^C}[NXN
	cTimer->setTimer( T_TOTAL_TIME );		// vJn

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// fBANeBx[g̏
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t Localization::onDeactivated(RTC::UniqueId ec_id){

	// vI
	cTimer->stopTimer( T_TOTAL_TIME );

	cout << "\n\nThe end of Localization.\n\n"
		 << "Total Execution Time : " << 0.001 * (double)cTimer->getValueAve(T_TOTAL_TIME)  << " [s]\n\n"
		 << "Required Time per cycle\n"
		 << "  Average : " <<  cTimer->getValueAve(T_ONE_CICLE) << " [ms/cycle]\n"
		 << "  Maximum : " <<  cTimer->getValueMax(T_ONE_CICLE) << " [ms]\n"
		 << "  Minimum : " <<  cTimer->getValueMin(T_ONE_CICLE) << " [ms]\n\n";

	delete input;
	delete kslam;
	delete cTimer;

	input = 0;
	kslam = 0;
	cTimer = 0;

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// s֐
//-------------------------------------------------------------------------------------------------+
RTC::ReturnCode_t Localization::onExecute(RTC::UniqueId ec_id){

	cTimer->setTimer( T_ONE_CICLE );

	InputPose();
	InputLRF();
	InputStereo();

	if( !input->IsNewPose() || !input->IsNewLRFData() ) return RTC::RTC_OK;

	kslam->Mapping();

	OutputPose();

	cTimer->stopTimer( T_ONE_CICLE );

	return RTC::RTC_OK;
}


//-------------------------------------------------------------------------------------------------+
// {bg̈ʒuEpf[^|[g󂯎
//-------------------------------------------------------------------------------------------------+
bool Localization::InputPose(){

	if( !m_RobotPoseIn.isNew() ) return false;

	while( !m_RobotPoseIn.isEmpty() ) m_RobotPoseIn.read();

	Pose pose;

	pose.x     = m_RobotPose.data.position.x * 1000;
	pose.y     = m_RobotPose.data.position.y * 1000;
	pose.angle = m_RobotPose.data.heading;

	input->SetData( pose );

	cerr << "Input Robot Pose\n";

	return true;
}


//-------------------------------------------------------------------------------------------------+
// [UZT̃f[^f[^|[g󂯎
//-------------------------------------------------------------------------------------------------+
bool Localization::InputLRF(){

	if( !m_RangeDataIn.isNew() ) return false;

	while( !m_RangeDataIn.isEmpty() ) m_RangeDataIn.read();

	SlamRangeData range;

	range.max_data  = m_RangeData.data.distance.length();
	range.min_angle = m_RangeData.data.startPosition -90.0;
	range.max_angle = m_RangeData.data.endPosition   -90.0;

	for( int i=0; i<range.max_data; i++ ){

		double value = m_RangeData.data.distance[i];

		if( MAX_RANGE < value || value < 10 ) value = -1;		// 擾\͈͓̃f[^mF

		range.value.push_back( value );
	}

	input->SetData( range );

	cerr << "Input LRF Data\n";

	return true;
}


//-------------------------------------------------------------------------------------------------+
//XeIJ̃f[^f[^|[g󂯎
//-------------------------------------------------------------------------------------------------+
bool Localization::InputStereo(){

	if( !m_ImageIn.isNew() || !m_StereoDataIn.isNew() ) return false;

	while( !m_ImageIn.isEmpty()      ) m_ImageIn.read();
	while( !m_StereoDataIn.isEmpty() ) m_StereoDataIn.read();

	input->SetData( TimedImageData2IplImage( &(m_Image)) );

	SlamStereoData stereo( m_StereoData.width, m_StereoData.height );

	for( int i=0; i<(int)m_StereoData.data.length(); i++ ){

		int col = ( i / (int)(m_StereoData.width) );
		int row = ( i % (int)(m_StereoData.width) );

		stereo.value[row][col][0] = -1000.0 * m_StereoData.data[i].x;
		stereo.value[row][col][1] =  1000.0 * m_StereoData.data[i].y;
		stereo.value[row][col][2] =  1000.0 * m_StereoData.data[i].z;
	}

	input->SetData( stereo );

	cerr << "Input Stereo Data\n";

	return true;
}


//-------------------------------------------------------------------------------------------------+
// {bg̈ʒuEpf[^|[gɏo͂
//-------------------------------------------------------------------------------------------------+
void Localization::OutputPose(){

	Pose pose = kslam->EstimatedPose();

	m_EstimatedPose.data.position.x = pose.x / 1000.0;
	m_EstimatedPose.data.position.y = pose.y / 1000.0;
	m_EstimatedPose.data.heading    = pose.angle;

	m_EstimatedPose.tm = m_RobotPose.tm;
	m_PosePair.tm      = m_RobotPose.tm;

	m_PosePair.ododata = m_RobotPose.data;
	m_PosePair.estdata = m_EstimatedPose.data;

	m_EstimatedPoseOut.write();
	m_PosePairOut.write();

	cerr << "Output Estimated Robot Pose ( "
		 << pose.x / 1000.0 << ", "
		 << pose.y / 1000.0 << ", "
		 << pose.angle << " )\n";
}


//-------------------------------------------------------------------------------------------------+
// 摜ϊ
//-------------------------------------------------------------------------------------------------+
IplImage* Localization::TimedImageData2IplImage( TimedImageData *data ){

	IplImage *img = cvCreateImage( cvSize( data->data.width,data->data.height ), data->data.depth, data->data.nChannels );
	for( int i=0; i<(int)data->data.imageSize; i++ ) img->imageData[i] = data->data.imgData[i];
	return img;
}


//-------------------------------------------------------------------------------------------------+
// T[rX|[g̐ڑmF
//-------------------------------------------------------------------------------------------------+
bool Localization::CheckServicePort(){

	if( CORBA::is_nil(m_AbsoluteMapService.getObject()) ){

		cerr << "Service port is not conected with service provider.\n";
		return false;
	}

	if( m_AbsoluteMapService.getObject()->_non_existent() ){
	
		cerr << "Service provider is not activated.\n";
		return false;
	}

	return true;
}


//-------------------------------------------------------------------------------------------------+
// n}ǂݍ
//-------------------------------------------------------------------------------------------------+
TimedFloatAbsoluteOGMapData *Localization::GetMapData(){

	OGMapConfig conf = m_AbsoluteMapService->getAbsoluteOGMapConfig();

	double x_scale = conf.xScale;
	double y_scale = conf.yScale;

	double origin_x = conf.origin.position.x;
	double origin_y = conf.origin.position.y;

	TimedFloatAbsoluteOGMapData *mapdata = m_AbsoluteMapService->getFloatAbsoluteOGMap( conf.origin.position.x, conf.origin.position.y, conf.width, conf.height );

	if( !mapdata || mapdata->cells.length() <= 0 ) return 0;

	return mapdata;
}


//-------------------------------------------------------------------------------------------------+
// n}ǂݍŏ
//-------------------------------------------------------------------------------------------------+
GlobalMap *Localization::InitGlobalMap(){

	while( !CheckServicePort() ) Sleep( 1000 );

	TimedFloatAbsoluteOGMapData *mapdata;

	while( !( mapdata = GetMapData() )){

		cerr << "GlobalMapData is not completed.\n";
		Sleep( 1000 );
	}

	MapDataForMCL g_map;

	if( m_USE_INIT_POSE ) g_map.SetInitialPose( Pose( m_INIT_X, m_INIT_Y, m_INIT_HEADING ));

	return g_map.GenerateGlobalMap( mapdata, m_Particle_num );
}


//**************************************************************************************************

//-------------------------------------------------------------------------------------------------+
// RXgN^
// ꎞIȑn}̃f[^̈m
//-------------------------------------------------------------------------------------------------+
MapDataForMCL::MapDataForMCL(){

	initial_pose = false;

	map.resize( G_MAP_WIDTH );

	for( int x=0; x<G_MAP_WIDTH; x++ ){

		map[x].resize( G_MAP_HEIGHT );

		for( int y=0; y<G_MAP_HEIGHT; y++ ){

			map[x][y].resize( G_MAP_size );

			for( int i=0; i<G_MAP_size; i++ ){

				map[x][y][i].resize( G_MAP_size, -1.0 );
			}
		}
	}
}


//-------------------------------------------------------------------------------------------------+
// W̎w
//-------------------------------------------------------------------------------------------------+
void MapDataForMCL::SetInitialPose( Pose pose ){

	initial_pose = true;
	this->pose = pose;

}


//-------------------------------------------------------------------------------------------------+
// n}f[^i[ĕԂ
//-------------------------------------------------------------------------------------------------+
GlobalMap *MapDataForMCL::GenerateGlobalMap( TimedFloatAbsoluteOGMapData *mapdata, int particle_num ){

	if( mapdata->mapconfig.xScale != G_grid_size / 1000.0 || mapdata->mapconfig.yScale != G_grid_size / 1000.0 ){

		cerr << "\n}bvXP[0.1m/cellȊO̒n}͌_ŖΉłI\n";
		throw;
	}

	double val;

	cerr << "map size = ( " << mapdata->mapconfig.width * mapdata->mapconfig.xScale << ", " << mapdata->mapconfig.height * mapdata->mapconfig.yScale << " );\n";

	for( int i=0; i<mapdata->cells.length(); i++ ){

		int x = i % mapdata->mapconfig.width + G_MAP_size/2;
		int y = i / mapdata->mapconfig.width + G_MAP_size/2;

		map[x/G_MAP_size + G_MAP_WIDTH/2][y/G_MAP_size + G_MAP_HEIGHT/2][x%G_MAP_size][y%G_MAP_size] = mapdata->cells[i];
	}

	// n}f[^i[
	Gmap_Piece *** input_map;
	Gmap_Piece* end_P = new Gmap_Piece();			//X^bNXg̍Ō|C^

	input_map = new Gmap_Piece** [G_MAP_WIDTH];

	for( int i = 0; i < G_MAP_WIDTH; i++ ) input_map[i] = new Gmap_Piece* [G_MAP_HEIGHT];

	bool input_map_flag[G_MAP_WIDTH][G_MAP_HEIGHT];

	for( int i=0; i<G_MAP_WIDTH; i++ ) for( int j=0; j<G_MAP_HEIGHT; j++ ){

		input_map_flag[i][j] = false;

		for( int x=0; x<G_MAP_size; x++ ) for( int y=0; y<G_MAP_size; y++ ){

			if( map[i][j][x][y] != -1.0 ){

				if( input_map_flag[i][j] == false ){

					input_map_flag[i][j] = true;

					Gmap_Piece* New_G = new Gmap_Piece();

					for( int ii=0; ii<G_MAP_size; ii++ ) for( int jj=0; jj<G_MAP_size; jj++ ) New_G->G_Piece[ii][jj] = -1;

					New_G->G_Piece[x][y] = map[i][j][x][y];
					input_map[i][j]      = New_G;
					end_P->Next          = New_G;
					New_G->Before        = end_P;
					end_P                = New_G;

					New_G->set_keeping_num(0);
				}

				else input_map[i][j]->G_Piece[x][y] = map[i][j][x][y];
			}
		}
	}

	int para[4];
	double start_x  = 0;
	double start_y  = 0;
	double start_th = 0;

	GlobalMap *GrobalMap_L = new GlobalMap[particle_num];

	for( int p=0; p<particle_num; p++ ){

		cerr << "particle[" << p << "] : ";

		// W̊i[
		if( initial_pose ){

			start_x  = pose.x * 1000.0;
			start_y  = pose.y * 1000.0;
			start_th = pose.angle;
		}
		else while(1){

			int i_x = rand()%mapdata->mapconfig.width;
			int i_y = rand()%mapdata->mapconfig.height;
			double value = mapdata->cells[ i_x + i_y*mapdata->mapconfig.width ];

			if( 0.3 <= value || value < 0.0 ) continue;

			start_x  = (double)i_x * mapdata->mapconfig.xScale * 1000.0;
			start_y  = (double)i_y * mapdata->mapconfig.yScale * 1000.0;
			start_th = ran0() * 2.0 * M_PI - M_PI;

			break;
		}

		cerr << "Pose( " << start_x/1000.0 << ", " << start_y/1000.0 << ", " << start_th << " )\n";

		GrobalMap_L[p].set_path( start_x, start_y, start_th, 0, 0 );

		for( int i=0; i<G_MAP_WIDTH; i++ ) for( int j=0; j<G_MAP_HEIGHT; j++ ){

			if( input_map_flag[i][j] == true ) GrobalMap_L[p].G_map_P( i, j, input_map[i][j] );
		}
	}

	return GrobalMap_L;
}



//-------------------------------------------------------------------------------------------------+
extern "C"
{
  void LocalizationInit(RTC::Manager* manager)
  {
    coil::Properties profile(localization_spec);
    manager->registerFactory(profile,
                             RTC::Create<Localization>,
                             RTC::Delete<Localization>);
  }  
};
