﻿//------------------------------------------------------------------------------------------------+
// Filename: GlobalMap.cpp
// Content : 大域地図関連クラスのメンバ関数
//------------------------------------------------------------------------------------------------+

#include "GlobalMap.h"
#include "edit_map.h"
#include <iostream>
using namespace std;


//------------------------------------------------------------------------------------------------+
// コンストラクタ
//------------------------------------------------------------------------------------------------+
GlobalMap::GlobalMap(){

	set_path( 0, 0, 0, 0, 0 );	//初期位置登録

	// 地図やフラグの初期化
	for( int i=0; i<G_MAP_WIDTH; i++) for( int j=0; j<G_MAP_HEIGHT; j++ ){

		G_map_N[i][j][0] = -1;
		G_map_N[i][j][1] = -1;
		Map_point[i][j]  = NULL;
	}

	likelihood     = 1.0;
	sum_likelihood = 1.0;
	likelihood_L   = 1.0;
	likelihood_G   = 1.0;
}


//------------------------------------------------------------------------------------------------+
// 経路のリストに追加(データは後で入力)
//------------------------------------------------------------------------------------------------+
void GlobalMap::make_new_path(){

	Position_class P;
	P.x  = 0;
	P.y  = 0;
	P.th = 0;
	path.push_back(P);
}


//------------------------------------------------------------------------------------------------+
// 引数はパーティクルの自己位置
// 引数で得た自己位置を経路リストに追加
//------------------------------------------------------------------------------------------------+
void GlobalMap::set_path( double x, double y, double th, double s, double ns ){

	make_new_path();	// リストの追加
	int age=get_age();	// 追加したリストの番号を得る

	// 情報の更新
	path[age-1].nsec=ns;
	path[age-1].sec=s;
	path[age-1].x=x;
	path[age-1].y=y;
	path[age-1].th=th;
}


//------------------------------------------------------------------------------------------------+
// ローカルマップを受け取って処理する
//------------------------------------------------------------------------------------------------+
void GlobalMap::input_lmap( LocalMap* localmap, double dis_x, double dis_y, double dis_th, double tth, double ku, Gmap_Piece** GP_end, Gmap_Piece** GP_re, int* GP_re_count ){

	int age = (int)( get_age() - 1 - LMap_charge_time );

	// パーティクルが持つ自己位置から大域地図中の自分がいるセルの位置を計算
	double sin_value = sin(( tth - path[age].th ));				//三角関数を先に計算
	double cos_value = cos(( tth - path[age].th ));				//三角関数を先に計算
	double change_x  =  cos_value * dis_x + sin_value * dis_y;	//オドメトリの移動量を相対座標から絶対座標へ
	double change_y  = -sin_value * dis_x + cos_value * dis_y;

	// 絶対座標を計算
	double world_x  = change_x + path[age].x;
	double world_y  = change_y + path[age].y;
	double world_th = dis_th   + path[age].th;

	// 対応グリッドを計算
	world_x = world_x / G_grid_size;
	world_y = world_y / G_grid_size;

	// 局所地回転用ダミーマップ
	static vector<vector<double>> dummy_map;
	if( dummy_map.empty() ) dummy_map = NewMapData( DUMMY_MAP_SIZE );
	else ResetMapData( &dummy_map );

	// 局所地図を回転させる
	DataRange data_range = TurnLocalMap( localmap, &dummy_map, world_th );

	// 局所地図を半分のサイズにする用のダミーマップ
	//（このためにセルサイズは大域地図は局所地図の2倍しか設定できない）
	static vector<vector<double>> dummy_map2;
	NewHalfSizeMap( &dummy_map2, &dummy_map, data_range );


	int para[4];
	int dis = 0;
	double dumyP;
	double d = 0;
	int count2 = 0;

	int relative_x = (int)( world_x + 0.5 * G_MAP_size * ( G_MAP_WIDTH  + 1.0 ));
	int relative_y = (int)( world_y + 0.5 * G_MAP_size * ( G_MAP_HEIGHT + 1.0 ));


	//グリッド地図の更新　////////////////////////////////////////
	for( int j= ( data_range.min_y - DUMMY_MAP_SIZE/2 ) / 2; j<( data_range.max_y - DUMMY_MAP_SIZE/2 )/2; j++ ){
		for( int i= ( data_range.min_x - DUMMY_MAP_SIZE/2 ) / 2; i<( data_range.max_x - DUMMY_MAP_SIZE/2 )/2; i++ ){

			// 入力の局所地図(絶対座標にあわせて回転させてセルサイズをあわせたもの)を1セルごと調べる

//			change_coordinates( world_x+i, world_y+j, para );

			para[0] = ( relative_x + i ) % G_MAP_size;
			para[1] = ( relative_y + j ) % G_MAP_size;

			para[2] = ( relative_x + i - para[0] ) / G_MAP_WIDTH;
			para[3] = ( relative_y + j - para[1] ) / G_MAP_HEIGHT;

			para[0] -= G_MAP_size   / 2;
			para[1] -= G_MAP_size   / 2;
			para[2] -= G_MAP_WIDTH  / 2;
			para[3] -= G_MAP_HEIGHT / 2;

			//局所地図のセルに対応する大域地図の地図パーツと
			// パーツ内の座標を計算する。
			// world_x+i,world_y+j に対応する大域地図のセルがpara[4]の中に入る
			// para[0],para[1]は地図パーツ内の座標、para[2],para[3]は地図パーツの地図の座標

			 /*地図の更新に際して地図パーツリストおよび地図パーツ地図の更新処理が入る


			 （この後のパーツの更新を行なうが、パーツが無い場合はパーツの作成しマップに登録、
			 他の閲覧者がいるパーツを更新するときはコピーを用意してそちらを更新するように
			 地図パーツ地図を変更する）

			 以下の３つの状態によってそれぞれ処理が変わる
			 ①このステップの更新で対象の地図パーツが1回も更新されていない
			 ②過去に対応する座標の地図パーツに更新が無い(無かった場合地図パーツが作られていない)
			 ③閲覧者がこのパーティクルだけである
			 ④再利用スタックに再利用待ちの地図パーツがない

			 処理は以下のように変化する。
			 －①T②T④Tの時
			 新たに地図パーツを作ってリストに追加
			 地図パーツ地図にはそのアドレスを登録
			 パーツの閲覧者を１にセット

			 －①T②T④Nの時
			 再利用スタックから地図パーツを取り出しデータを初期化
			 地図パーツ地図にはそのパーツのアドレスを登録
			 パーツの閲覧者を１にセット

			 －①T②N③Tの時
			 　何もしない

			  －①T②N③N④Tの時
			  地図パーツ地図が示すパーツの閲覧者を一人減らす
			  新たに地図パーツを作ってリストに追加,地図パーツ地図が示すパーツの情報をコピーする
			  地図パーツ地図にはコピーしたパーツのアドレスを登録


			  －①T②N③N④Nの時
			  地図が示すパーツの閲覧者を一人減らす
			  再利用スタックから地図パーツを取り出しデータを初期化,地図パーツ地図が示すパーツの情報をコピーする
			  地図パーツ地図にはコピーしたパーツのアドレスを登録

			  －①N
			  何もしない
			  */

			if( G_map_N[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ][1] < 0 ){ 

				if( Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ] == NULL ){

					if( *GP_re_count == 0 ){

						//－①T②T④Tの時
						Gmap_Piece* New_G = NULL;
						New_G = new Gmap_Piece();

						for( int i=0; i<G_MAP_size; i++ ) for( int j=0; j<G_MAP_size; j++ ) New_G->G_Piece[i][j] = -1;

						Map_point[para[2] + (G_MAP_HEIGHT/2)][para[3] + (G_MAP_WIDTH/2)] = New_G;
						(*GP_end)->Next = New_G;
						New_G->Before   = (*GP_end);
						(*GP_end)       = New_G;

						G_map_N[para[2] + (G_MAP_HEIGHT/2)][para[3]+(G_MAP_WIDTH/2)][1] = 1;
						New_G->set_keeping_num(1);
					}
					else{
						//①T②T④Nの時
						for( int i=0; i<G_MAP_size; i++ ) for( int j=0; j<G_MAP_size; j++ ){
							GP_re[(*GP_re_count)-1]->G_Piece[i][j] = -1;
						}

						Map_point[para[2] + (G_MAP_HEIGHT/2)][para[3] + (G_MAP_WIDTH/2)] = GP_re[(*GP_re_count)-1];
						(*GP_end)->Next = GP_re[(*GP_re_count)-1];
						GP_re[(*GP_re_count)-1]->Before = (*GP_end);
						(*GP_end) = GP_re[(*GP_re_count)-1];

						G_map_N[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ][1] = 1;
						GP_re[ (*GP_re_count) - 1 ]->set_keeping_num(1);
						(*GP_re_count)--;
					}
				}
				else{ 
					//①T②N③Tの時
					if( Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ]->keeping_num == 1 ){
						G_map_N[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ][1] = 1;
					}
					else{
						if( *GP_re_count == 0 ){
							//①T②N③N④Tの時
							Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ]->sub_keeping_num();
							Gmap_Piece* New_G = NULL;
							New_G = new Gmap_Piece();
							Map_point[para[2] + (G_MAP_HEIGHT/2)][para[3]+(G_MAP_WIDTH/2)]->copy( New_G );

							Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ] = New_G; 
							G_map_N[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2)][1 ] = 1;
							New_G->set_keeping_num(1);

							(*GP_end)->Next = New_G;
							New_G->Before = (*GP_end);
							(*GP_end) = New_G;
						}
						else{
							//①T②N③N④Nの時
							Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ]->sub_keeping_num();
							Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ]->copy( GP_re[(*GP_re_count-1)] );

							Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ] = GP_re[(*GP_re_count)-1]; 
							G_map_N[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ][1] = 1;
							GP_re[(*GP_re_count)-1]->set_keeping_num(1);

							(*GP_end)->Next = GP_re[(*GP_re_count)-1];
							GP_re[(*GP_re_count)-1]->Before = (*GP_end);
							(*GP_end) = GP_re[(*GP_re_count-1)];
							(*GP_re_count)--;
						}
					}
				}
			}


			// 尤度の計算用にdis とcount2を計算
			if( 0 <= Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ]->G_Piece[ para[0] + (G_MAP_size/2) ][ para[1] + (G_MAP_size/2) ] ){

				if( 0 <= dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] ){

					if( !(( dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ]>0.49 ) && ( dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] < 0.51 ))){
					
						if( Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2)]->G_Piece[ para[0] + (G_MAP_size/2) ][ para[1] + (G_MAP_size/2) ] >= 0.8 ){

							if( dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] < 0.3 ) dis++;	//物体があるという観測に対応したセルが占有状態である場合の数
							count2++;	//物体があるという観測の数
						}
					}
				}
			}


			// セルの更新
/*			double dummy = 0;

			if( Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ]->G_Piece[ para[0] + (G_MAP_size/2) ][ para[1] + (G_MAP_size/2) ] < 0 ){

				//更新セルば未観測状態の時
				if( 0 <= dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] ){

					//物体があるという観測の時
					if( dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] >= 0.7 && dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] <= 1.0 ){
						
						dummy = ( poe(d) * 0.5 ) / ( poe(d) * 0.5 + poe_(d) * ( 1 - 0.5 ));

						//確率の上限(=1)にならないようにする
						if( dummy > 0.999 )	dummy = 0.999;

						Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ]->G_Piece[para[0] + (G_MAP_size/2) ][ para[1] + (G_MAP_size/2) ] = dummy;		//確率の更新
					}

					//物体がないという観測の時
					else if( dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] <= 0.2 && dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] >= 0 ){
					
						dummy = ( po_e(d) * 0.5 ) / ( po_e(d) * 0.5 + po_e_(d) * ( 1 - 0.5 )); 

						//確率の下限(=0)にならないようにする
						if( dummy < 0.001 ) dummy = 0.001;

						Map_point[ para[2] + (G_MAP_HEIGHT / 2) ][ para[3] + (G_MAP_WIDTH/2) ]->G_Piece[ para[0] + (G_MAP_size/2) ][ para[1] + (G_MAP_size/2) ] = dummy;		//確率の更新
					}
				}
			}

			else{
				//更新セルば未観測状態以外のとき
				if( 0 <= dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] ){
					if(!(( dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] > 0.49 ) && ( dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] < 0.51 ))){

						dumyP = Map_point[ para[2] + ( G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2)]->G_Piece[ para[0] + (G_MAP_size/2) ][ para[1] + (G_MAP_size/2) ];		//更新前のセルの確率
 
						if( dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] >= 0.7 ){

							//物体があるという観測の時
							dummy = ( poe(d) * dumyP ) / ( poe(d) * dumyP + poe_(d) * ( 1 - dumyP ));

							//確率の上限(=1)にならないようにする
							if( dummy > 0.999 )	dummy = 0.999;

							Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ]->G_Piece[para[0] + (G_MAP_size/2) ][ para[1] + (G_MAP_size/2) ] = dummy;		//確率の更新
						}

						else if( dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] <= 0.2 && dummy_map2[ i + DUMMY_MAP_SIZE/4 ][ j + DUMMY_MAP_SIZE/4 ] >= 0 ){

							//物体がないという観測の時
							dummy = ( po_e(d) * dumyP ) / ( po_e(d) * dumyP + po_e_(d) * ( 1 - dumyP ));

							//確率の下限(=0)にならないようにする
							if( dummy < 0.001 ) dummy = 0.001;

							Map_point[ para[2] + (G_MAP_HEIGHT/2) ][ para[3] + (G_MAP_WIDTH/2) ]->G_Piece[ para[0] + (G_MAP_size/2) ][ para[1] + (G_MAP_size/2) ] = dummy;		//確率の更新
						}
					}
				}
			}*/
		}
	}


	// 尤度の計算
	double kappa = ku;
	if( count2 > 0 ) likelihood_G *= exp( -kappa * (double)dis / (double)count2 );
	else likelihood_G *= 0.5;

	// セルの値に応じて尤度を下げる
	if( GetCellValue() < 0.0 || 0.5 <= GetCellValue() ) likelihood_G = 0.0;
	else if( 0.3 <= GetCellValue() ) likelihood_G *= 0.1;

	// パーツの更新フラグのリセット
	for( int j=0; j<G_MAP_HEIGHT; j++ ) for( int i=0; i<G_MAP_WIDTH; i++ ) G_map_N[i][j][1] = -1;
}


//------------------------------------------------------------------------------------------------+
// 回転用のダミーマップのメモリ領域を確保
//------------------------------------------------------------------------------------------------+
vector<vector<double>> GlobalMap::NewMapData( int map_size ){

	vector<vector<double>> dummy_map;
	dummy_map.resize( map_size );

	for( int i=0; i<map_size; i++ ) dummy_map[i].resize( map_size, -1.0 );

	return dummy_map;
}


//------------------------------------------------------------------------------------------------+
// 回転用のダミーマップをリセットする
//------------------------------------------------------------------------------------------------+
void GlobalMap::ResetMapData( vector<vector<double>> *map ){

	for( int i=0; i<(int)(*map).size(); i++ ) for( int j=0; j<(int)(*map)[i].size(); j++ ){

		(*map)[i][j] = -1.0;
	}
}


//------------------------------------------------------------------------------------------------+
// 局所地図を回転させる
// ついでに情報が入っているグリッドの範囲を調べる
//------------------------------------------------------------------------------------------------+
DataRange GlobalMap::TurnLocalMap( LocalMap* localmap, vector<vector<double>> *dummy_map, double theta ){

	int max_x = 0;
	int max_y = 0;
	int min_x = DUMMY_MAP_SIZE;
	int min_y = DUMMY_MAP_SIZE;

	double sin_value = sin( theta );
	double cos_value = cos( theta );

	 for( int j=-DUMMY_MAP_SIZE / 2; j<DUMMY_MAP_SIZE / 2; j++ ){
		 for( int i=-DUMMY_MAP_SIZE/2; i<DUMMY_MAP_SIZE / 2; i++ ){

			 double dumy_x =  cos_value * i + sin_value * j;
			 double dumy_y = -sin_value * i + cos_value * j;

			 if( dumy_y <= (LOCALMAP_HEIGHT/2-1) && dumy_y >= -(LOCALMAP_HEIGHT/2+1) && dumy_x <= (LOCALMAP_WIDTH/2-1) && dumy_x >= -(LOCALMAP_WIDTH/2+1)){

				 // 回転前のグリッドの座標の計算　4近傍の値を見る
				 double odd_x = fabs( dumy_x - (int)dumy_x );
				 double odd_y = fabs( dumy_y - (int)dumy_y );

				 double dummy;

				 dummy  = (1-odd_x) * (1-odd_y) * localmap->get_atlas( (int)(dumy_x+(LOCALMAP_WIDTH/2  )), (int)((LOCALMAP_HEIGHT/2)+dumy_y  ) );
				 dummy += ( odd_x ) * (1-odd_y) * localmap->get_atlas( (int)(dumy_x+(LOCALMAP_WIDTH/2+1)), (int)((LOCALMAP_HEIGHT/2)+dumy_y  ) );
				 dummy += ( odd_x ) * ( odd_y ) * localmap->get_atlas( (int)(dumy_x+(LOCALMAP_WIDTH/2+1)), (int)((LOCALMAP_HEIGHT/2)+dumy_y+1) );
				 dummy += (1-odd_x) * ( odd_y ) * localmap->get_atlas( (int)(dumy_x+(LOCALMAP_WIDTH/2  )), (int)((LOCALMAP_HEIGHT/2)+dumy_y+1) );

				(*dummy_map)[ i + DUMMY_MAP_SIZE / 2 ][ j + DUMMY_MAP_SIZE / 2 ] = dummy;	// 値の代入

				// 次のステップ以降で全てのグリッドを見なくてすむように情報が入っているグリッドの範囲を調べる
				if( !( dummy > 0.45 && dummy < 0.55 )){

					if( max_x < i + DUMMY_MAP_SIZE / 2 ) max_x = i + DUMMY_MAP_SIZE / 2;
					if( max_y < j + DUMMY_MAP_SIZE / 2 ) max_y = j + DUMMY_MAP_SIZE / 2;
					if( min_x > i + DUMMY_MAP_SIZE / 2 ) min_x = i + DUMMY_MAP_SIZE / 2;
					if( min_y > j + DUMMY_MAP_SIZE / 2 ) min_y = j + DUMMY_MAP_SIZE / 2;
				}
			}
			else (*dummy_map)[ i + DUMMY_MAP_SIZE / 2 ][ j + DUMMY_MAP_SIZE / 2 ] = -1.0;
		}
	}

	DataRange data;
	data.max_x = max_x;
	data.max_y = max_y;
	data.min_x = min_x;
	data.min_y = min_y;

	return AdjustRange( data );
}


//------------------------------------------------------------------------------------------------+
// グリッドの情報がある範囲を調整(少し大きめに範囲をとる)
//------------------------------------------------------------------------------------------------+
DataRange GlobalMap::AdjustRange( DataRange data ){

	if( data.max_y < DUMMY_MAP_SIZE - 5 ) data.max_y += 5;
	if( data.min_y > 5 ) data.min_y -= 5;
	if( data.max_x < DUMMY_MAP_SIZE - 5 ) data.max_x += 5;
	if( data.min_x > 5 ) data.min_x -= 5;

	if( ( data.max_y - data.min_y )%2 == 1 ){

		if( data.min_y == 0 ) data.max_y++;
		else data.min_y--;
	}

	if( ( data.max_x - data.min_x )%2 == 1 ){

		if( data.min_x == 0 ) data.max_x++;
		else data.min_x--;
	}

	return data;
}


//------------------------------------------------------------------------------------------------+
// ダミーマップを縦横半分の大きさにした地図を生成して返す
// つまり、局所地図と大域地図のセルサイズをあわせる
//------------------------------------------------------------------------------------------------+
void GlobalMap::NewHalfSizeMap( vector<vector<double>> *dummy_map2, vector<vector<double>> *dummy_map, DataRange data ){

	if( dummy_map2->empty() ) *dummy_map2 = NewMapData( DUMMY_MAP_SIZE/2 );
	else ResetMapData( dummy_map2 );

	for( int j=data.min_y; j<data.max_y; j+=2 ){
		for( int i=data.min_x; i<data.max_x; i+=2 ){

			int flag = 0;
			double max_dumy = 0;

			// まとめられる4セルの中で最大値をそのセルの確率とする
			if( 0 <= (*dummy_map)[i][j] ){

				if( max_dumy < (*dummy_map)[i][j] ) max_dumy = (*dummy_map)[i][j];
				flag++;
			}
			if( 0 <= (*dummy_map)[i+1][j] ){

				if( max_dumy < (*dummy_map)[i+1][j] ) max_dumy = (*dummy_map)[i+1][j];
				flag++;
			}
			if( 0 <= (*dummy_map)[i+1][j+1] ){

				if( max_dumy < (*dummy_map)[i+1][j+1] ) max_dumy = (*dummy_map)[i+1][j+1];
				flag++;
			}
			if( 0 <= (*dummy_map)[i][j+1] ){

				if( max_dumy < (*dummy_map)[i][j+1] ) max_dumy = (*dummy_map)[i][j+1];
				flag++;
			}
			if( flag == 0 ) (*dummy_map2)[i/2][j/2]=-1;		// 4セルとも情報がない場合は－1
			else{
				 (*dummy_map2)[i/2][j/2] = max_dumy;			// 4セル中最大の確率とする

				 if( (*dummy_map2)[i/2][j/2] < 0.5 && (*dummy_map2)[i/2][j/2] > 0.49 ) (*dummy_map2)[i/2][j/2] = 0.51;
			}
		}
	}
}


//------------------------------------------------------------------------------------------------+
// パーティクルのコピー
//------------------------------------------------------------------------------------------------+
void GlobalMap::copy_data( GlobalMap original ){

	sum_likelihood = original.get_sum_likelihood();		// 尤度
	path = original.path;		// 経路

	// グリッド地図のコピー
	for( int j=0; j<G_MAP_HEIGHT; j++ ) for( int i=0; i<G_MAP_WIDTH; i++ ){

		// 地図パーツ地図の初期化＋閲覧者－1(コピー先のパーティクルは消えるため)
		if( Map_point[i][j] != NULL ){

			Map_point[i][j]->sub_keeping_num();
			Map_point[i][j] = NULL;
		}

		G_map_N[i][j][1] = -1;		//地図パーツ地図更新フラグのリセット

		// 地図パーツ地図の更新（コピー元と同じポインタを登録）＋コピー元の地図パーツの閲覧者+1
		if( original.Map_point[i][j] != NULL ){

			Map_point[i][j] = original.Map_point[i][j];
			original.Map_point[i][j]->add_keeping_num();
		}
	}

	likelihood_G = original.get_likelihood_G();	//尤度　グリッド
	likelihood_L = original.get_likelihood_L();	//尤度	ランドマーク
	lmark = original.lmark;						//ランドマーク地図
}


//------------------------------------------------------------------------------------------------+
// グリッドマップの描画
//------------------------------------------------------------------------------------------------+
void GlobalMap::show_grid_map( IplImage* img ){

	for( int i=0; i<G_MAP_WIDTH; i++ ) for( int j=0; j<G_MAP_HEIGHT; j++ ){ 

		if( Map_point[i][j] != NULL ){

			// 地図パーツがある場合
			for( int y=0; y < G_MAP_size; y++ ) for( int x=0; x < G_MAP_size; x++ ){

				int GD, gggg;

				//セルが未観測
				if( Map_point[i][j]->G_Piece[x][y] == -1 ){
					GD = 100;
					gggg = 0;
				}

				//セルが観測
				else{
					GD = (int)( 0.5 + 255 * Map_point[i][j]->G_Piece[x][y] );
					gggg = 1;
				}

				// セルの座標を表示画像の座標に変更(SOP_xとSOP_yは位置調整用)
				int show_y = SOP_y - ( x + G_MAP_size * ( i - G_MAP_WIDTH/2)  ) + G_MAP_size/2;
				int show_x = SOP_x - ( y + G_MAP_size * ( j - G_MAP_HEIGHT/2) ) + G_MAP_size/2;

				if(show_x>=0&&show_x<img->width&&show_y>=0&&show_y<img->height){

					if( gggg == 1 ){
						// 観測なら0から1の確率を0から255に変換
						(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3] = GD;
						(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3 + 1] = GD;
						(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3 + 2] = GD;
					}
					else{
						// 未観測のセルは緑に
						(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3] = 255;
						(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3 + 1] = 100;
						(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3 + 2] = 100;
					}
				}
			}
		}
	}
}


//------------------------------------------------------------------------------------------------+
// 大域地図画像を作成
//------------------------------------------------------------------------------------------------+
IplImage* GlobalMap::GlobalMapImage(){

	// 画像サイズの決定
	DataRange size = GlobalMapSize();

	int width  = G_MAP_size * ( size.max_x - size.min_x + 1 );
	int height = G_MAP_size * ( size.max_y - size.min_y + 1 );

	if( width <= 0 || height <= 0 ) return 0;

	IplImage* img = cvCreateImage( cvSize( width, height ), IPL_DEPTH_8U, 3);

	// 各画素に値を格納する
	for( int i=size.min_x; i<=size.max_x; i++ ) for( int j=size.min_y; j<=size.max_y; j++ ){ 

		for( int y=0; y < G_MAP_size; y++ ) for( int x=0; x < G_MAP_size; x++ ){

			// 描画する座標を計算
			int show_x =            x + G_MAP_size * ( i - size.min_x );
			int show_y = height - ( y + G_MAP_size * ( j - size.min_y )) -1;

			//セルが未観測
			if( Map_point[i][j] == NULL || Map_point[i][j]->G_Piece[x][y] == -1 ){

				(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3]     = 255;
				(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3 + 1] = 100;
				(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3 + 2] = 100;
			}

			// 観測があれば0から1の確率を0から255に変換して値を格納
			else{
				int value = (int)( 0.5 + 255.0 * Map_point[i][j]->G_Piece[x][y] );

				(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3]     = value;
				(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3 + 1] = value;
				(img)->imageData[(img)->widthStep * (show_y) + (show_x) * 3 + 2] = value;
			}
		}
	}

	return img;
}


//------------------------------------------------------------------------------------------------+
// 大域地図画像サイズ(範囲)の決定
//------------------------------------------------------------------------------------------------+
DataRange GlobalMap::GlobalMapSize(){

	DataRange size;
	size.min_x = G_MAP_WIDTH;
	size.min_y = G_MAP_HEIGHT;

	for( int i=0; i<G_MAP_WIDTH; i++ ) for( int j=0; j<G_MAP_HEIGHT; j++ ){

		if( Map_point[i][j] != NULL ){

			if( size.max_x < i ) size.max_x = i;
			if( i < size.min_x ) size.min_x = i;
			if( size.max_y < j ) size.max_y = j;
			if( j < size.min_y ) size.min_y = j;
		}
	}

	return size;
}


//------------------------------------------------------------------------------------------------+
// マップクリーニング用の訪問地図cfを埋めるための再帰関数
//
// 引数がセルの座標で　この関数が呼ばれたらcfの対応場所を1にする
// また、このセルが占有状態以外で未観測で無い場合、周囲8近傍を呼び出し再帰を行なう
//------------------------------------------------------------------------------------------------+
void GlobalMap::make_cf( int i, int j, int x, int y ){

	if(i<G_MAP_WIDTH&&j<G_MAP_HEIGHT){
		if(x<0){
			i--;
			x=G_MAP_size-1;
		}
		if(y<0){
			j--;
			y=G_MAP_size-1;
		}

		if(x>=G_MAP_size){
			i++;
			x=0;
		}
		if(y>=G_MAP_size){
			j++;
			y=0;
		}
		//printf("%d %d %d %d \n",i,j,x,y);


		if(!(Map_point[i][j]==NULL)&&cf[i][j][x][y]==0){
			cf[i][j][x][y]=1;
			if((Map_point[i][j]->G_Piece[x][y]<0.7)&&!(Map_point[i][j]->G_Piece[x][y]==-1)){
				make_cf(i,j,x-1,y-1);
				make_cf(i,j,x,y-1);
				make_cf(i,j,x+1,y-1);
				make_cf(i,j,x-1,y);
				make_cf(i,j,x+1,y);
				make_cf(i,j,x-1,y+1);
				make_cf(i,j,x,y+1);
				make_cf(i,j,x+1,y+1);

			}
		}
	}
}


//------------------------------------------------------------------------------------------------+
// 地図クリーニング　あまり意味が無いかも　見た目がきれいになるだけ　現在未使用
//
// 再帰関数をつかって、スタート地点からみて占有以外のセルに隣接しているセルのを調べる
// 上記のセル以外を消すことで、不要に厚くなった壁や誤観測のセルを消せる
//
// 訪問地図cf を作成初期化　このcfが0のセルは未訪問　1は訪問
// 再帰を使って、スタート地点からみて占有以外のセルに隣接しているセルのcfを1にしていく
// (自分の8近傍周辺を調べていく)
//------------------------------------------------------------------------------------------------+
void GlobalMap::clean_map(Gmap_Piece** GP_end,Gmap_Piece** GP_re,int* GP_re_count){


	int flag[G_MAP_WIDTH][G_MAP_HEIGHT];
	cf =new int***[G_MAP_WIDTH];
	for(int i=0;i<G_MAP_WIDTH;i++){
		cf[i]= new int**[G_MAP_HEIGHT];
	}
	for(int i=0;i<G_MAP_WIDTH;i++){
		for(int j=0;j<G_MAP_HEIGHT;j++){
			cf[i][j]= new int*[G_MAP_size];
		}
	}
	for(int i=0;i<G_MAP_WIDTH;i++){
		for(int j=0;j<G_MAP_HEIGHT;j++){
			for(int x=0;x<G_MAP_size;x++){
				cf[i][j][x]= new int[G_MAP_size];
			}
		}
	}

	for(int i=0;i<G_MAP_WIDTH;i++){
		for(int j=0;j<G_MAP_HEIGHT;j++){

			flag[i][j] = 0;

			for(int x=0;x<G_MAP_size;x++){
				for(int y=0;y<G_MAP_size;y++){
					cf[i][j][x][y]=0;
				}
			}
		}
	}


	// セルのチェックy
	make_cf( G_MAP_WIDTH/2, G_MAP_HEIGHT/2, G_MAP_size/2, G_MAP_size/2 );



	// cfが1のセル(スタート地点からみて占有以外のセルに隣接しているセル)以外のセルを消す
	for( int i=0; i<G_MAP_WIDTH; i++ ){ 
		for( int j=0; j<G_MAP_HEIGHT; j++ ){ 
			if( Map_point[i][j] != NULL ){
				for( int y=0; y < G_MAP_size; y++ ){
					for( int x=0; x < G_MAP_size; x++ ){        
						if( cf[i][j][x][y] == 0 ){
							if( flag[i][j] == 1 ){
								Map_point[i][j]->G_Piece[x][y] = -1;
							}
							else{
								if( Map_point[i][j]->keeping_num == 1 );

								else if((*GP_re_count)==0){
									Map_point[i][j]->sub_keeping_num();
									Gmap_Piece* New_G=NULL;
									New_G =new Gmap_Piece();
									Map_point[i][j]->copy(New_G);

									Map_point[i][j]=New_G; 

									New_G->set_keeping_num(1);

									(*GP_end)->Next=New_G;
									New_G->Before=(*GP_end);
									(*GP_end)=New_G;
								}
								else{
									// printf("BBBBBBBB\n");
									Map_point[i][j]->sub_keeping_num();
									Map_point[i][j]->copy(GP_re[(*GP_re_count-1)]);

									Map_point[i][j]=GP_re[(*GP_re_count)-1]; 
									GP_re[(*GP_re_count)-1]->set_keeping_num(1);

									(*GP_end)->Next=GP_re[(*GP_re_count)-1];
									GP_re[(*GP_re_count)-1]->Before=(*GP_end);
									(*GP_end)=GP_re[(*GP_re_count-1)];
									(*GP_re_count)--;

								}

								flag[i][j]=1;
								Map_point[i][j]->G_Piece[x][y]=-1;
							}
						}
					}
				}
			}
		}
	}

	for(int i=0;i<G_MAP_WIDTH;i++){
		for(int j=0;j<G_MAP_HEIGHT;j++){
			for(int x=0;x<G_MAP_size;x++){
				delete[] cf[i][j][x];
			}
			delete[] cf[i][j];

		}
		delete[] cf[i];
	}
	delete[] cf;

}


//------------------------------------------------------------------------------------------------+
// グリッド地図の情報をテキストに書き出す
//------------------------------------------------------------------------------------------------+
void GlobalMap::write_grid_map(FILE* fp){

	for( int i=0; i<G_MAP_WIDTH; i++ ){ 
		for( int j=0; j<G_MAP_HEIGHT; j++ ){ 
			if( Map_point[i][j] != NULL ){

				for( int y=0; y < G_MAP_size; y++ ){
					for( int x=0; x < G_MAP_size; x++ ){        
						if( Map_point[i][j]->G_Piece[x][y] != -1 ){

							fprintf(fp,"%d\t%d\t%d\t%d\t%lf\n",i,j,x,y,Map_point[i][j]->G_Piece[x][y]);
						}	      	      
					}
				}
			}
		}
	}
}


//------------------------------------------------------------------------------------------------+
// 最新の自己位置を結果画像に丸で表示
//------------------------------------------------------------------------------------------------+
void GlobalMap::show_pose( IplImage* img, int R,int B,int G ){

	int i = get_age()-1;

	int x = (int)( path[i].x / G_grid_size );
	int y = (int)( path[i].y / G_grid_size );

	CvScalar rcolor = CV_RGB ( R, B, G );

	CvPoint pt1= cvPoint( SOP_x - y, SOP_y - x );

	cvCircle( img, pt1, 1,rcolor );
}


//------------------------------------------------------------------------------------------------+
// ランドマーク地図を表示
// 第3引数はdefineで定義された画像サイズと入力された画像サイズの比
//------------------------------------------------------------------------------------------------+
void GlobalMap::show_Imark( IplImage* img, vector<z_struct_3> Z_S, double resize ){

	// Z_Sは対応が取れた観測(SIFT特徴)とランドマークの情報が入っている
	for( int i=0;i<(get_lmark_num());i++){

		CvScalar rcolor = CV_RGB (255,255,255);

		CvPoint pt1; // ランドマークのポイント
		CvPoint pt2; // 画像上のSIFT特徴　pt1とpt2は対応が取れている。

		pt2.x = SOP_x - (int)( lmark[i].myu_y / G_grid_size );
		pt2.y = SOP_y - (int)( lmark[i].myu_x / G_grid_size );

		if( lmark[i].mach_z != -1 ){

			int num = lmark[i].mach_z;

			pt1.x = resize * Z_S[num].window_x / 2;
			pt1.y = resize * Z_S[num].window_y / 2;

			arrow( img, pt1, pt2, rcolor, 10, 1, 8, 0 ); // 対応が取れた観測とランドマークを白い線で結ぶ
		}


		// カルマンフィルタで計算したランドマークの範囲を楕円で表示
		rcolor = CV_RGB (0,255,0);
		CvSize axex;
		double ramda1,ramda2;
		double a,b,c,d;//共分散

		a = ( lmark[i].var_11 );
		b = ( lmark[i].var_21 );
		c = ( lmark[i].var_12 );
		d = ( lmark[i].var_22 );

		ramda1 = ( (a+b) +sqrt( (a+d)*(a+d) -4*( a*d -c*b ))) / 2.0;
		ramda2 = ( (a+b) -sqrt( (a+d)*(a+d) -4*( a*d -c*b ))) / 2.0;

		double daen_y = 1;
		double daen_x = b / ( ramda1 - a );

		axex.width = (int)( sqrt( fabs( ramda1 ))) / G_grid_size;

		daen_y = 1;
		daen_x = b / ( ramda2 - a );
		//double daen_th=atan2(daen_y,daen_x)*180.0/3.14;

		axex.height = (int)( sqrt( fabs( ramda2 )))/G_grid_size;

		if( axex.width  == 0 ) axex.width  = 1;
		if( axex.height == 0 ) axex.height = 1;

		rcolor = CV_RGB( 0, 100, 255 );

		if( lmark[i].nsec > 0 ){
			//cvEllipse( img,pt2,axex,daen_th,0.0, 360.0, rcolor);		//分散を楕円で表字
		}


		// ランドマークの表示
		if( lmark[i].nsec > 4 ){				//nsecにはランドマークの更新回数が入っている
			rcolor = CV_RGB( 150, 255, 0 );
			cvCircle( img, pt2, 1, rcolor );	//4回以上観測されたランドマークを表示
		}

	}
}


//------------------------------------------------------------------------------------------------+
// スタートから今までの経路を表示する
//------------------------------------------------------------------------------------------------+
void GlobalMap::show_path(IplImage* img, int R,int B,int G){

	for( int i=0; i<(int)(get_age()-1); i++ ){

		int x = (int)(path[i].x/G_grid_size);
		int y = (int)(path[i].y/G_grid_size);
		int show_y, show_x;

		show_x = SOP_x - y;
		show_y = SOP_y - x;

		CvScalar rcolor = CV_RGB (R,B,G);

		CvPoint pt1;
		pt1.x = show_x;
		pt1.y = show_y;

		cvCircle( img, pt1, 1, rcolor );
	}
}



//------------------------------------------------------------------------------------------------+
// ランドマークを新規作成
//------------------------------------------------------------------------------------------------+
void GlobalMap::make_new_lmark(){

	// 新たなランドマークを作る
	Land_Mark LM;
	LM.myu_x=0.0;
	LM.myu_y=0.0;
	LM.myu_z=0.0;

	LM.var_11=0.0;
	LM.var_12=0.0;
	LM.var_13=0.0;

	LM.var_21=0.0;
	LM.var_22=0.0;
	LM.var_23=0.0;

	LM.var_31=0.0;
	LM.var_32=0.0;
	LM.var_33=0.0;

	for( int i=0; i<128; i++ ) LM.SIFT_Val[i]=0;

	LM.sec  = 0.0;
	LM.nsec = 0.0;
	LM.discovery_degree = 0;

	for( int l=0; l<100; l++ ){

		LM.observation_time[l]=0;
		LM.discovery_degree_S[l]=-1;
	}

	LM.discovery_num = 0;

	lmark.push_back(LM);	//ベクターに登録
}


//------------------------------------------------------------------------------------------------+
//寿命によって消えるランドマークのチェック
//
// 引数のtimeは現在のステップ数、ランドマークが最後に観測数
// そのためtime-lmark[i].secはlmark[i]が最後に観測されてからの経過数となる
// nsecは更新回数(ただし、初回更新が0)
//------------------------------------------------------------------------------------------------+
void GlobalMap::erase_lmark(double time){

	int age = get_lmark_num();

	for( int i=0; i<age; i++ ){

		if( lmark[i].nsec < 2 ){
			double past_time = time-lmark[i].sec;
			if(past_time>LIMIT_TIME){
				destroy_lmark(i);			//対象のランドマークをリストから削除
				i--;
			}
		}
		else if( lmark[i].nsec < 6 ){
			double past_time = time-lmark[i].sec;
			if( past_time > ( LIMIT_TIME * 50 )){
				destroy_lmark(i);			//対象のランドマークをリストから削除
				i--;
			}
		}
		age = get_lmark_num();
	}
}


//------------------------------------------------------------------------------------------------+
// 対象のランドマークを削除する
//------------------------------------------------------------------------------------------------+
void GlobalMap::destroy_lmark(int i){

	lmark.erase(lmark.begin()+i); 
}


//------------------------------------------------------------------------------------------------+
// ランドマークのコピー(引数のランドマークがコピー元)
//------------------------------------------------------------------------------------------------+
void GlobalMap::set_new_lmark(Land_Mark LM){

	make_new_lmark();//ランドマークリストの最後に新規のランドマークを作成
	int age=get_lmark_num();

	// コピー元のランドマークの情報を新規ランドマークにコピー
	lmark[age-1].myu_x=LM.myu_x;
	lmark[age-1].myu_y=LM.myu_y;
	lmark[age-1].myu_z=LM.myu_z;

	lmark[age-1].var_11=LM.var_11;
	lmark[age-1].var_12=LM.var_12;
	lmark[age-1].var_13=LM.var_13;

	lmark[age-1].var_21=LM.var_21;
	lmark[age-1].var_22=LM.var_22;
	lmark[age-1].var_23=LM.var_23;

	lmark[age-1].var_31=LM.var_31;
	lmark[age-1].var_32=LM.var_32;
	lmark[age-1].var_33=LM.var_33;
	lmark[age-1].discovery_degree=LM.discovery_degree;

	for(int i=0;i<128;i++){
		lmark[age-1].SIFT_Val[i]=LM.SIFT_Val[i];
	}
	lmark[age-1].sec=LM.sec;
	lmark[age-1].nsec=LM.nsec;
	lmark[age-1].discovery_num=LM.discovery_num;

	for(int l=0;l<0;l++){
		lmark[age-1].observation_time[l]=LM.observation_time[l];
		lmark[age-1].discovery_degree_S[l]=LM.discovery_degree_S[l];
	}
}


//------------------------------------------------------------------------------------------------+
// ランドマークの更新
//------------------------------------------------------------------------------------------------+
void GlobalMap::input_z( vector<z_struct_3> Z_S, double s, double ns ){

	int data_num = (int) Z_S.size();

	Land_Mark New;
	matrix Q;
	clock_t step_s,step_e;
	double one=0,two=0,three=0;
	step_s=clock();

	int age=get_age();
	double R_th=path[age-1].th;
	double R_x=path[age-1].x;
	double R_y=path[age-1].y;
	double cos_value=cos(R_th);
	double sin_value=sin(R_th);  
	
	vector<double> wx;
	vector<double> wy;
	vector<int> match_K;

	wx.resize( data_num );
	wy.resize( data_num );
	match_K.resize( data_num );

	int invalid_num=0;


	// ステレオ距離情報の観測はロボットを基準とした相対座標なので
	// スタート地点を基準とした絶対座標に変換
	for(int i=0;i<data_num;i++){
		wx[i]=cos_value*Z_S[i].z_x-sin_value*Z_S[i].z_y+R_x;	//絶対座標x
		wy[i]=sin_value*Z_S[i].z_x+cos_value*Z_S[i].z_y+R_y;	//絶対座標y
		match_K[i]=-1;
	}

	for( int j=0; j<(get_lmark_num()); j++ ) lmark[j].mach_z = -1;

	double average_min_distance = 0.0;

	step_e=clock();
	one=(double)(step_e-step_s)/CLOCKS_PER_SEC;

	vector<int> use_lmark;
	vector<vector<double>> use_lmark_val;
	vector<vector<double>> use_lmark_Kval;
	use_lmark.resize( (int)get_lmark_num() );
	use_lmark_val.resize( (int)get_lmark_num() );
	use_lmark_Kval.resize( (int)get_lmark_num() );

	for( int i=0; i<(int)get_lmark_num(); i++ ){
		use_lmark_val[i].resize(2);
		use_lmark_Kval[i].resize(2);
	}

	//ランドマークがセンサ範囲かどうかを見るフラグ:tureの時は範囲内 
	vector<bool> EFF_lmark;
	EFF_lmark.resize( (int)get_lmark_num() );

	for( int j=0; j<(get_lmark_num()); j++){

		use_lmark[j] = -1;
		use_lmark_val[j][0]  = -1;
		use_lmark_val[j][1]  = -1;
		use_lmark_Kval[j][0] = -1;
		use_lmark_Kval[j][1] = -1;
		EFF_lmark[j] = false;

		// パーティクルの自己位置からランドマークの相対位置を計算する
		// センサ範囲と照合してランドマーク無いか調べる
		double dx = lmark[j].myu_x - R_x;
		double dy = lmark[j].myu_y - R_y;
		double karichi = sqrt( sq(dx) + sq(dy) ); 

		//センサ範囲かどうかチェック(距離)
		if( karichi < 8000.0 && karichi > 1500.0 ){

			double theta = atan2( dy, dx );

			if( fabs( theta - R_th ) < ( 50 * M_PI / 180.0 )
				&& fabs( theta - lmark[j].discovery_degree ) < ( 20 * M_PI / 180.0 )){	// センサ範囲かどうかチェック(角度範囲)
					EFF_lmark[j] = true;	// センサ無いのランドマークはtrueとなる		
			}
		}
	}

	int count = 0;

	for( int i=0; i<data_num; i++ ){
		double min_distance = -1;
		int min_distance_num = -1;
		double second_E_val = -1;
		double first_E_val = -1;

		for( int j=0; j<(get_lmark_num()); j++ ){    
			// 観測に対してランドマークを全比較

			if(fabs(lmark[j].myu_x-wx[i])<1000
				&&fabs(lmark[j].myu_y-wy[i])<1000
				&&fabs(lmark[j].myu_z-Z_S[i].z_z)<1000){//ランドマークと観測の絶対距離の条件(あまり離れているのは対応させない)
				
					if(EFF_lmark[j]==true){//ランドマークがセンサの範囲内 
					double E_val = 0;
					count++;

					for( int k=0; k<128; k++ ){
						
						//特徴量の差を計算
						E_val += pow( lmark[j].SIFT_Val[k] - Z_S[i].SIFT_Val[k], 2 );
					}  

					// 全比較のうちもっとも近い時の差を保存
					// 観測に対してランドマークの全比較を行なっている
					// 例え、観測に近いランドマークがあっても同じくらい近いランドマークがあると
					// 誤対応のもとになるので特徴量の差が近い1番目と2番目のを比較する
					if( first_E_val == -1 ) first_E_val=E_val;
					else if( first_E_val > E_val ){		//2回目以降をは1番近い差と2番目目に近い時の差を保存

						second_E_val = first_E_val;
						first_E_val = E_val;
					}

					if( E_val < K_SIFT_VAL ){		//特徴量の差がしきい値未満かどうか

						double distance=check_Mahalanobis(j,wx[i],wy[i]);//マハラノビス距離のチェック

						// ここまでチェックを通ったランドマークの中でマハラノビス距離が一番近い
						//　ランドマークの番号とマハラノビス距離を記憶
						if( min_distance_num == (-1) ){		//min_distance_num==(-1)の時は条件初通過
							min_distance = distance;
							min_distance_num = j;
						}
						else if( min_distance > distance ){
							min_distance = distance;
							min_distance_num = j;
						}
					}
				}
			}
		}


		
		if( second_E_val != -1 ){

			//!（second_E_val==-1）の場合は観測とランドマークが複数とれた場合
			if( fabs( second_E_val - first_E_val ) < ( SecondB_val * SecondB_val )){

				// 対応が取れた複数のランドマークの中に似たようなランドマークがあった時
				// 処理：その観測は使わない

				invalid_num++;	// 使わなかった観測の数
				match_K[i]=-2;	// -2は観測を使わないことを意味する
			}
			else{ 
				// 対応が取れた複数のランドマークの中に似たようなランドマークが無かった時
				if(!(min_distance==-1)&&(min_distance<9.0)){	//マハラノビス距離が9の時暫定的にマッチング成立

					// マッチングが取れたランドマークが他の観測ともマッチングをとったか調べる
					if(use_lmark[min_distance_num]==-1){
						// この観測が初めてのマッチングの場合

						// マハラノビス距離を最短距離として記憶
						use_lmark_val[min_distance_num][1]=use_lmark_val[min_distance_num][0];//2番目に近いマハラノビス距離
						use_lmark_val[min_distance_num][0]=min_distance;
						use_lmark[min_distance_num]=i;//ランドマークが対応をとった観測を記憶する
													  //この跡により近い観測が無ければマッチングが確定
			
						average_min_distance+=pow(min_distance,1);//尤度計算用にマッチングが取れた時のマハラノビス距離の合計を計算する
						match_K[i]=min_distance_num;//マッチングが取り消された場合↑の値を合計から引くために値を保存
						 

					}
					else if(use_lmark_val[min_distance_num][0]>min_distance){
						// 2つ以上の観測が同じランドマークをマッチングがとられた場合
						// かつ、以前にマッチングがとられた観測よりマハラノビス距離が近い時は
						// 前のマッチングを取り消して、新しいマッチングを採用

						average_min_distance-=pow(use_lmark_val[min_distance_num][0],1); 
						match_K[use_lmark[min_distance_num]]=-2;
						match_K[i]=min_distance_num;
						use_lmark_val[min_distance_num][1]=use_lmark_val[min_distance_num][0];
						use_lmark_val[min_distance_num][0]=min_distance;
						use_lmark[min_distance_num]=i;  
						invalid_num++;
					}
					else{
						// 2つ以上マッチングが取れたが、以前のマッチングの方が近かった場合
						invalid_num++;
						match_K[i]=-2;
					}
				}
				else{
					// マハラノビス距離が9以上の場合はマッチングをとらない
					average_min_distance+=pow(9.0,1);	// 尤度計算のマハラノビス距離の合計値は＋9とする
				}
			}
		}
		else {
			if(!(min_distance==-1)&&(min_distance<9.0)){

				if(use_lmark[min_distance_num]==-1){
					use_lmark_val[min_distance_num][1]=use_lmark_val[min_distance_num][0];
					

					use_lmark_val[min_distance_num][0]=min_distance;
					use_lmark_Kval[min_distance_num][1]=use_lmark_Kval[min_distance_num][0];
					use_lmark_Kval[min_distance_num][0]=first_E_val;

					use_lmark[min_distance_num]=i;      
					match_K[i]=min_distance_num;
					average_min_distance+=pow(min_distance,1); 
				}
				else if(use_lmark_Kval[min_distance_num][0]>first_E_val){
					average_min_distance-=pow(use_lmark_val[min_distance_num][0],1); 

					match_K[use_lmark[min_distance_num]]=-2;
					match_K[i]=min_distance_num; 
					use_lmark_val[min_distance_num][1]=use_lmark_val[min_distance_num][0];

					use_lmark_Kval[min_distance_num][1]=use_lmark_Kval[min_distance_num][0];
					use_lmark_val[min_distance_num][0]=min_distance;
					use_lmark_Kval[min_distance_num][0]=first_E_val;


					use_lmark[min_distance_num]=i;  
					invalid_num++;
				}
				else{
					invalid_num++;
					match_K[i]=-2;
				}

			}
			else{
				average_min_distance+=pow(9.0,1);
			}
		}

	}

	// マッチングが複数取れた場合
	for( int j=0; j<(get_lmark_num()); j++ ) if( use_lmark_val[j][1] != -1 ){

		average_min_distance -= pow( use_lmark_val[j][0], 1 ); 
		match_K[use_lmark[j]] =- 2;
		invalid_num++;
	}


	// 追加　　更新回数が4回以下ランドマークの場合は尤度の計算を除外する
	int carinum = 0;
	for( int i=0; i<data_num; i++ ) if( match_K[i]!=-2 && match_K[i] !=-1 ){

		if(lmark[match_K[i]].nsec<3){
			//4回以下のランドマークの分のマハラノビス距離を引く
			average_min_distance-=pow(use_lmark_val[match_K[i]][0],1); 
			invalid_num++;//この時の数を数えておく
		}
		else{
			carinum++;
		}
	}

	step_e = clock();

	two = (double)(step_e-step_s) / CLOCKS_PER_SEC;
	double kappa = 5.0;

	likelihood_L *= exp( -kappa * ( 1.0 - ( 9.0 * ( data_num - invalid_num ) - average_min_distance ) / ( 9.0 * ( data_num - invalid_num ) + 0.1 )));


	for(int i=0;i<data_num;i++){
		if(match_K[i]==-2){
		}
		else if(match_K[i]==-1){

			Q=get_Q_mat(wx[i],wy[i]);

			New.var_11=Q.S00;
			New.var_12=Q.S01;
			New.var_21=Q.S10;
			New.var_22=Q.S11;
			New.var_13=0.0; 
			New.var_23=0.0;
			New.var_31=0.0;
			New.var_32=0.0;
			New.var_33=0.0;
			New.myu_x=wx[i];
			New.myu_y=wy[i];
			New.myu_z=Z_S[i].z_z;
			double dx=wx[i]-R_x;
			double dy=wy[i]-R_y;
			double Theta=atan2(dy,dx);
			New.discovery_degree=Theta;

			for(int k=0;k<128;k++){
				New.SIFT_Val[k]=Z_S[i].SIFT_Val[k];
			}

			New.sec=s;
			New.nsec=ns;
			New.discovery_num=1;
			for(int l=0;l<100;l++){
				New.observation_time[l]=0;
				New. discovery_degree_S[l]=-1;
			}
			New.observation_time[0]=(int)(s);
			New.discovery_degree_S[0]=Theta;
			New.discovery_num=1;
			set_new_lmark(New);

		} 
		else{


			update_lmark(match_K[i],wx[i],wy[i],Z_S[i].SIFT_Val,s,ns);

			for(int k=0;k<128;k++){
				lmark[match_K[i]].SIFT_Val[k]=
					(double)(lmark[match_K[i]].SIFT_Val[k]*lmark[match_K[i]].nsec+Z_S[i].SIFT_Val[k])/
					(double)(lmark[match_K[i]].nsec+1.);
			}



			lmark[match_K[i]].mach_z=i; 
			lmark[match_K[i]].myu_z=(Z_S[i].z_z+lmark[match_K[i]].myu_z*lmark[match_K[i]].nsec)/(1+lmark[match_K[i]].nsec);
			double dx=wx[i]-R_x;
			double dy=wy[i]-R_y;
			double Theta=atan2(dy,dx);

			lmark[match_K[i]].discovery_degree
				=(double)(Theta+lmark[match_K[i]].discovery_degree*lmark[match_K[i]].nsec)/(double)(1+lmark[match_K[i]].nsec);
			while(lmark[match_K[i]].discovery_degree>2*M_PI){
				lmark[match_K[i]].discovery_degree-=2*M_PI;
			}
			while(lmark[match_K[i]].discovery_degree<-2*M_PI){
				lmark[match_K[i]].discovery_degree+=2*M_PI;
			}


			if((lmark[match_K[i]].discovery_num)<100){

				lmark[match_K[i]].discovery_degree_S[lmark[match_K[i]].discovery_num]=Theta;
				lmark[match_K[i]].observation_time[lmark[match_K[i]].discovery_num]=(int)(s);
				lmark[match_K[i]].discovery_num=(int)(lmark[match_K[i]].discovery_num)+1;
			}
		}
	}

	erase_lmark(s);

	step_e=clock();
	three=(double)(step_e-step_s)/CLOCKS_PER_SEC;
}


//------------------------------------------------------------------------------------------------+
//------------------------------------------------------------------------------------------------+
void GlobalMap:: update_lmark(int lmark_ID,double x,double y,double z,double* SI_Val,double s,double ns){

	lmark[lmark_ID].sec=s;

	matrix K;
	matrix SIG_V;
	SIG_V.S00 = lmark[lmark_ID].var_11;
	SIG_V.S10 = lmark[lmark_ID].var_21;
	SIG_V.S01 = lmark[lmark_ID].var_12;
	SIG_V.S11 = lmark[lmark_ID].var_22;


	matrix Q;
	Q = get_Q_mat( lmark[lmark_ID].myu_x, lmark[lmark_ID].myu_y );

	matrix dummy;	//S+Q
	dummy.S00 = SIG_V.S00 + Q.S00;
	dummy.S10 = SIG_V.S10 + Q.S10;
	dummy.S01 = SIG_V.S01 + Q.S01;
	dummy.S11 = SIG_V.S11 + Q.S11;

	matrix dummy2;
	dummy2.S00 = 0.0;
	dummy2.S10 = 0.0;
	dummy2.S01 = 0.0;
	dummy2.S11 = 0.0; 

	// (S+Q)^-1
	double D = ( dummy.S00 * dummy.S11 - dummy.S10 * dummy.S01 );

	if( D != 0 ){
		dummy2.S00 =  dummy.S11 / D;
		dummy2.S10 = -dummy.S10 / D;
		dummy2.S01 = -dummy.S01 / D;
		dummy2.S11 =  dummy.S00 / D; 
	}
	else printf("CL ERROR");

	K = aaa( SIG_V, dummy2 );


	double dumy3[1][2];
	dumy3[0][0] = x - lmark[lmark_ID].myu_x;
	dumy3[0][1] = y - lmark[lmark_ID].myu_y;

	lmark[lmark_ID].myu_x += K.S00 * dumy3[0][0] + K.S10 * dumy3[0][1];
	lmark[lmark_ID].myu_y += K.S01 * dumy3[0][0] + K.S11 * dumy3[0][1];

	dummy2.S00 = 0.0;
	dummy2.S10 = 0.0;
	dummy2.S01 = 0.0;
	dummy2.S11 = 0.0; 

	dummy.S00 = 0.0;
	dummy.S10 = 0.0;
	dummy.S01 = 0.0;
	dummy.S11 = 0.0; 

	dummy.S00 = 1.0 - K.S00;
	dummy.S10 = -K.S10;
	dummy.S01 = -K.S01;
	dummy.S11 = 1.0 - K.S11;

	dummy2 = aaa( dummy, SIG_V );

	lmark[lmark_ID].var_11 = dummy2.S00;
	lmark[lmark_ID].var_21 = dummy2.S10;
	lmark[lmark_ID].var_12 = dummy2.S01;
	lmark[lmark_ID].var_22 = dummy2.S11;
	lmark[lmark_ID].nsec++;
}


//------------------------------------------------------------------------------------------------+
//------------------------------------------------------------------------------------------------+
matrix GlobalMap::get_Q_mat( double myu_x, double myu_y ){

	matrix Q;
	int age = get_age();

	double R_th = path[age-1].th;
	double R_x = path[age-1].x;
	double R_y = path[age-1].y;
	double cos_value = cos( R_th );
	double sin_value = sin( R_th );
	double myu_x_R =  cos_value * ( myu_x - R_x ) + sin_value * ( myu_y - R_y );
	double myu_y_R = -sin_value * ( myu_x - R_x ) + cos_value * ( myu_y - R_y );

	double ur = F_VAL * (myu_x_R-B_VAL)/myu_y_R;
	double ul = F_VAL * (myu_x_R+B_VAL)/myu_y_R;

	if( ur - ul == 0 ) ur += 0.000001;

	double Delta = 2 * B_VAL / ((ul-ur)*(ul-ur));

	matrix A_T;
	matrix A;
	A.S00=-1*ur*Delta;
	A.S01=-1*F_VAL*Delta;
	A.S10=ul*Delta;
	A.S11=F_VAL*Delta;



	A_T.S00=A.S00;
	A_T.S01=A.S10;
	A_T.S10=A.S01;
	A_T.S11=A.S11;



	matrix S_img;
	S_img.S00=0.1;
	S_img.S01=0.0;
	S_img.S10=0.0;
	S_img.S11=0.1;

	matrix dummy,dummy2;
	dummy=aaa(A, S_img);

	dummy2=aaa(dummy,A_T);



	matrix R;
	R.S00=cos_value;
	R.S01=-sin_value;
	R.S10=sin_value;
	R.S11=cos_value;



	matrix R_T;
	R_T.S00=cos_value;
	R_T.S01=sin_value;
	R_T.S10=-sin_value;
	R_T.S11=cos_value;



	dummy=aaa(R,dummy2);
	Q=aaa(dummy,R_T);


	Q.S00=Q.S00/1.0;
	Q.S01=Q.S01/1.0;
	Q.S10=Q.S10/1.0;
	Q.S11=Q.S11/1.0;
	return Q;


}


//------------------------------------------------------------------------------------------------+
//------------------------------------------------------------------------------------------------+
void GlobalMap:: update_lmark(int lmark_ID,double x,double y,double* SI_Val,double s,double ns){

	lmark[lmark_ID].sec=s;
	//lmark[lmark_ID].nsec=ns;

	matrix K;
	matrix SIG_V;
	SIG_V.S00 = lmark[lmark_ID].var_11;
	SIG_V.S10 = lmark[lmark_ID].var_21;
	SIG_V.S01 = lmark[lmark_ID].var_12;
	SIG_V.S11 = lmark[lmark_ID].var_22;

	matrix Q;
	Q = get_Q_mat(lmark[lmark_ID].myu_x,lmark[lmark_ID].myu_y);

	matrix dummy;		//S+Q
	dummy.S00 = SIG_V.S00 + Q.S00;
	dummy.S10 = SIG_V.S10 + Q.S10;
	dummy.S01 = SIG_V.S01 + Q.S01;
	dummy.S11 = SIG_V.S11 + Q.S11;

	matrix dummy2;
	dummy2.S00 = 0.0;
	dummy2.S10 = 0.0;
	dummy2.S01 = 0.0;
	dummy2.S11 = 0.0; 


	// (S+Q)^-1
	double D = ( dummy.S00 * dummy.S11 - dummy.S10 * dummy.S01 );

	if( D !=0 ){
		dummy2.S00 =  dummy.S11 / D;
		dummy2.S10 = -dummy.S10 / D;
		dummy2.S01 = -dummy.S01 / D;
		dummy2.S11 =  dummy.S00 / D; 
	}
	else printf("CL ERROR");

	K = aaa( SIG_V, dummy2 );


	double dumy3[1][2];
	dumy3[0][0] = x - lmark[lmark_ID].myu_x;
	dumy3[0][1] = y - lmark[lmark_ID].myu_y;


	lmark[lmark_ID].myu_x += K.S00 * dumy3[0][0] + K.S10*dumy3[0][1];
	lmark[lmark_ID].myu_y += K.S01 * dumy3[0][0] + K.S11*dumy3[0][1];


	dummy2.S00 = 0.0;
	dummy2.S10 = 0.0;
	dummy2.S01 = 0.0;
	dummy2.S11 = 0.0; 

	dummy.S00 = 0.0;
	dummy.S10 = 0.0;
	dummy.S01 = 0.0;
	dummy.S11 = 0.0; 

	dummy.S00 = 1.0 - K.S00;
	dummy.S10 = -K.S10;
	dummy.S01 = -K.S01;
	dummy.S11 = 1.0 - K.S11;

	dummy2 = aaa( dummy, SIG_V );

	lmark[lmark_ID].var_11 = dummy2.S00;
	lmark[lmark_ID].var_21 = dummy2.S10;
	lmark[lmark_ID].var_12 = dummy2.S01;
	lmark[lmark_ID].var_22 = dummy2.S11;
	lmark[lmark_ID].nsec++;

}



//------------------------------------------------------------------------------------------------+
//------------------------------------------------------------------------------------------------+
double GlobalMap::check_Mahalanobis( int RM_ID, double x, double y ){

	double k = 100.0;

	double delta = lmark[RM_ID].var_11 * lmark[RM_ID].var_22 - lmark[RM_ID].var_12 * lmark[RM_ID].var_21;

	if( delta != 0 ){

		double sb_11 = ( lmark[RM_ID].var_22 ) / delta;
		double sb_12 = (-lmark[RM_ID].var_12 ) / delta;
		double sb_21 = (-lmark[RM_ID].var_21 ) / delta;
		double sb_22 = ( lmark[RM_ID].var_11 ) / delta;
		double A = ( x-lmark[RM_ID].myu_x );
		double B = ( y-lmark[RM_ID].myu_y );
		k = sqrt( A * A * sb_11 + A * B * ( sb_12 + sb_21 ) + B * B * sb_22 );
	}
	return k;
}


//------------------------------------------------------------------------------------------------+
// 移動量を入力
//------------------------------------------------------------------------------------------------+
void GlobalMap::set_path( double dis_x, double dis_y, double dis_th, double tth ){

	int age = (int)(get_age()-1);

	double sin_value2 = sin((tth-path[age].th));
	double cos_value2 = cos((tth-path[age].th));
	double change_x =  cos_value2 * dis_x + sin_value2 * dis_y;
	double change_y = -sin_value2 * dis_x + cos_value2 * dis_y;
	double world_x  = change_x + path[age].x;
	double world_y  = change_y + path[age].y;
	double world_th = dis_th+path[age].th;

	while( world_th >  M_PI ) world_th -= 2 * M_PI;
	while( world_th < -M_PI ) world_th += 2 * M_PI;

	set_path( world_x, world_y, world_th, 0, 0 );
}


//------------------------------------------------------------------------------------------------+
// 自己位置推定用？
//------------------------------------------------------------------------------------------------+
void GlobalMap::G_map_P( int i, int j, Gmap_Piece *piece ){

	G_map_N[i][j][1] = 1;
	Map_point[i][j]  = piece;

	Map_point[i][j]->add_keeping_num();
}


//------------------------------------------------------------------------------------------------+
// 大域地図の指定座標での障害物の占有確立を返す
// 地図の外を指定すると-1を返す
//------------------------------------------------------------------------------------------------+
double GlobalMap::GetCellValue( double dx, double dy ){

	int p = get_age()-1;

	double p_x = path[p].x + dx;
	double p_y = path[p].y + dy;

	int relative_x = (int)( p_x / G_grid_size + 0.5 * G_MAP_size * ( G_MAP_WIDTH  + 1.0 ));
	int relative_y = (int)( p_y / G_grid_size + 0.5 * G_MAP_size * ( G_MAP_HEIGHT + 1.0 ));

	int x = relative_x % G_MAP_size;
	int y = relative_y % G_MAP_size;

	int i = ( relative_x - x ) / G_MAP_WIDTH;
	int j = ( relative_y - y ) / G_MAP_HEIGHT;

	// 地図パーツがあるか
	if( x < 0 || G_MAP_WIDTH <= x || y < 0 || G_MAP_HEIGHT <= y || Map_point[i][j] == NULL ) return -1.0;

	return Map_point[i][j]->G_Piece[x][y];
}
