﻿//------------------------------------------------------------------------------------------------+
// Filename: GlobalMap.h
// Content : 大域地図関連クラスの定義
//------------------------------------------------------------------------------------------------+
#ifndef _G_MAP
#define _G_MAP

#include <stdio.h>
#include "cv.h"
#include "highgui.h"
#include <stdlib.h>
#include <time.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <algorithm>
#include "LocalMap.h"
#include "landmark.h"

#include <vector>
#include <iostream>


//------------------------------------------------------------------------------------------------+
// マクロの定義
//------------------------------------------------------------------------------------------------+
#define G_grid_size		(100)	// 大域グリッド地図のセルサイズ[mm]
#define G_MAP_WIDTH		(50)	// 地図パーツの横の個数
#define G_MAP_HEIGHT	(50)	// 地図パーツの縦の個数
#define G_MAP_size		(50)	// 地図パーツ一つのグリッド数
#define DUMMY_MAP_SIZE	(300)	// 入力の局所地図サイズ
#define LMap_charge_time (1)
#define SOP_x			(400)
#define SOP_y			(200)


//------------------------------------------------------------------------------------------------+
// 大域地図の1つの部分（ピース）を表すクラス
//------------------------------------------------------------------------------------------------+
class Gmap_Piece{

public:
	Gmap_Piece* Next;			// リストの次へのポインタ
	Gmap_Piece* Before;			// リストの前へのポインタ
	double G_Piece[G_MAP_size][G_MAP_size];	// 地図パーツのグリッド情報
	int keeping_num;			// 閲覧数
	int ID;

	Gmap_Piece(){
		keeping_num = 0;
		ID = -1;
		Next = NULL;
		Before = NULL;
	}


	// コピーコンストラクタ
	Gmap_Piece(const Gmap_Piece &src){

		keeping_num=src.keeping_num;
		for(int i=0;i<G_MAP_size;i++){
			for(int j=0;j<G_MAP_size;j++){
				G_Piece[i][j]=src.G_Piece[i][j];
			}
		}
	}


	// 地図パーツの初期化(再利用用)
	void new_Piece(){

		keeping_num = 1;
		for( int i=0; i<G_MAP_size; i++ ) for( int j=0; j<G_MAP_size; j++ ) G_Piece[i][j]=-1;
	}


	void add_keeping_num(){keeping_num++;}		//閲覧数を増やす
	void sub_keeping_num(){keeping_num--;}		//閲覧数を減らす
	void set_keeping_num(int i){keeping_num=i;}	//閲覧数をセット


	// 引数で渡された地図をコピーする
	void copy( Gmap_Piece *src ){ 

		for( int i=0; i<G_MAP_size; i++ ) for( int j=0; j<G_MAP_size; j++ ){

			src->G_Piece[i][j] = -1;
			src->G_Piece[i][j] = G_Piece[i][j];
		}
	}

};


//------------------------------------------------------------------------------------------------+
// 自己位置記録用のクラス
// パーティクル(ロボット)の位置
// これをベクターでつないで経路を表す
//------------------------------------------------------------------------------------------------+
class Position_class{
public:

	double x;
	double y;
	double th;
	double sec;
	double nsec;
};


//------------------------------------------------------------------------------------------------+
// データの入っている範囲を表す
//------------------------------------------------------------------------------------------------+
class DataRange{
public:
	int max_x;
	int max_y;
	int min_x;
	int min_y;

	DataRange(){ max_x = max_y = min_x = min_y = 0; }
};


//------------------------------------------------------------------------------------------------+
// 大域地図クラス
// パーティクルとして使う
//------------------------------------------------------------------------------------------------+
class GlobalMap{ 
private:
	double sum_likelihood;		// 累積尤度

public:
	GlobalMap();
	int G_map_N[G_MAP_WIDTH][G_MAP_HEIGHT][2];			// 地図パーツの管理用(更新チェック(1回の観測では1回更新可能)と使用チェック(地図があるか))
	Gmap_Piece*  Map_point[G_MAP_WIDTH][G_MAP_HEIGHT];	// 地図パーツのポインタ地図(地図パーツの位置関係)
	double likelihood_L;		// ランドマーク地図尤度
	double likelihood_G;		// グリッド地図尤度
	double likelihood;			// 全体尤度
	std::vector<Land_Mark> lmark;			// ランドマーク地図
	std::vector<Position_class> path;		// 経路情報(推定位置系列)


	// 自己位置関係
	void make_new_path();		//経路リストに一つデータを付け加える
	void set_path( double x, double y, double th, double s, double ns );		//自己位置をセットする
	int get_age(){ return path.size(); }						//経路の長さ≈ステップ数を返す
	void show_path( IplImage* img, int R, int G, int B );		//経路を表示
	void show_pose( IplImage* img, int R, int B, int G );		//自己位置を表示
	void set_path( double dis_x, double dis_y, double dis_th, double tth );		//自己位置をセットする


	// グリッド地図関係
	void input_lmap( LocalMap* localmap,double disx,double disy,double disth, double tth,double ka,Gmap_Piece** GP_end,Gmap_Piece** GP_re,int* GP_re_count);	//大域地図を更新
	vector<vector<double>> NewMapData( int map_size );		// 回転用のダミーマップのメモリ領域を確保
	void ResetMapData( vector<vector<double>> *map );		// 回転用のダミーマップをリセットする
	DataRange TurnLocalMap( LocalMap* localmap, vector<vector<double>> *dummy_map, double theta );		// 局所地図を回転させる
	DataRange AdjustRange( DataRange data );		// グリッドの情報がある範囲を調整
	void NewHalfSizeMap( vector<vector<double>> *dummy_map2, vector<vector<double>> *dummy_map, DataRange data );			// ダミーマップを縦横半分の大きさにした地図を生成して返す
	void copy_data(GlobalMap original);				//パーティクルのコピー
	void copy_Gmap(GlobalMap original,int i,int j);	//グリッド地図のコピー
	void write_grid_map(FILE* fp);					//グリッド地図の書き出し(テキスト形式)
	void clean_map(Gmap_Piece** GP_end,Gmap_Piece** GP_re,int* GP_re_count);		//グリッド地図をきれいにする
	int**** cf;		//clean_map関数で使う再起呼び出しで訪問したかどうか記述する地図
	void make_cf(int i,int j,int x,int y);		//再起探索を行い訪問地図をつくる
	void show_grid_map(IplImage* img);			// グリッド地図の画像を作成
	IplImage* GlobalMapImage();					// 大域地図画像を作成
	DataRange GlobalMapSize();					// 大域地図画像サイズ(範囲)の決定


	// 尤度のセット・取得
	double get_likelihood(){ return likelihood; }
	double get_sum_likelihood(){ return sum_likelihood; }
	void set_likelihood( double x ){ likelihood = x; }
	void set_sum_likelihood( double x ){ sum_likelihood = x; }


	// グリッド地図の更新に使う確率
	double poe(   double d ){ return 0.6; }
	double poe_(  double d ){ return 0.2; }
	double po_e(  double d ){ return 1-poe(d); }
	double po_e_( double d ){ return (1-poe_(d)); }


	// ランドマーク地図関連
	void make_new_lmark();		//新しいランドマークを作る
	void set_new_lmark(Land_Mark LM);		//新しいランドマークをリストに追加
	int get_lmark_num(){return lmark.size();}		//ランドマークのリストの長さを返す（ランドマークの数）
	void update_lmark(int lmark_ID,double x,double y,double z,double* SI_Val,double s,double ns);		//ランドマークの更新
	void input_z( vector<z_struct_3> Z_S, double s, double ns );
	void destroy_lmark(int i);		//ランドマークをリストから削除
	void erase_lmark(double time);		//ランドマークの時間制限のチェック　条件に合ったものを削除
	void update_lmark(int lmark_ID,double x,double y,double* SI_Val,double s,double ns);		//ランドマークの更新
	double check_Mahalanobis( int RM_ID,double x,double y);		//ランドマークと観測の間のマハラノビス距離を計算
	matrix get_Q_mat(double myu_x,double myu_y);		//初期観測誤差の取得　カルマンフィルタ用　ステレオカメラの観測モデル採用
	void show_Imark(IplImage* img, vector<z_struct_3> Z_S, double resize=1.0);	//ランドマーク地図の画像を作成する

	// 自己位置推定用？
	void G_map_P( int i, int j, Gmap_Piece* GP );

	// 尤度のセット・取得
	double get_likelihood_L(){ return likelihood_L; }
	void set_likelihood_L(double x){ likelihood_L = x; }
	double get_likelihood_G(){ return likelihood_G; }
	void set_likelihood_G(double x){ likelihood_G = x; }

	// For Localization
	double GetCellValue( double dx=0.0, double dy=0.0 );
};


#endif 
