﻿//------------------------------------------------------------------------------------------------+
// Filename: landmark.h
// Content : ランドマーククラスの定義
//------------------------------------------------------------------------------------------------+

#ifndef _LAND_MARK
#define _LAND_MARK


#include <stdio.h>
#include "cv.h"
#include "highgui.h"
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <algorithm>
#include <vector>
#include <iostream>


//------------------------------------------------------------------------------------------------+
// マクロを定義
//------------------------------------------------------------------------------------------------+
#define IMG_WIDTH	(512)
#define IMG_HEIGHT	(384)
#define Iblue(i,x,y) ((uchar*)((i)->imageData + (i)->widthStep*(y)))[3*(x)]		// 24ビットカラー、青
#define Igreen(i,x,y) ((uchar*)((i)->imageData + (i)->widthStep*(y)))[3*(x)+1]	// 24ビットカラー、緑
#define Ired(i,x,y) ((uchar*)((i)->imageData + (i)->widthStep*(y)))[3*(x)+2]	// 24ビットカラー、赤
#define K_SIFT_VAL	(250.0*250.0)	// SIFT特徴の特徴量の差のしきい値
#define SecondB_val (250)			// SIFT特徴の特徴量の差のしきい値
#define B_VAL		(120.0)			// ステレオカメラのカメラ間隔
#define F_VAL		(6.0)			// SIFT特徴の焦点距離
#define LIMIT_TIME	(3)				// ランドマークの寿命


//------------------------------------------------------------------------------------------------+
// 小さいほうの値を返す
//------------------------------------------------------------------------------------------------+
inline unsigned char min(int x, int y){
	return ( x < y )? x: y;
}


//------------------------------------------------------------------------------------------------+
// ランドマーククラス
//------------------------------------------------------------------------------------------------+
class Land_Mark{
public:

	//ランドマーク位置
	double  myu_x,  myu_y,  myu_z;
	double var_11, var_12, var_13;
	double var_21, var_22, var_23;
	double var_31, var_32, var_33;

	double SIFT_Val[128];	//SIFT特徴
	double sec, nsec;

	int observation_time[100];			//観測ステップ　対応状況確認用
	double discovery_degree;			// 観測時の方向
	double discovery_degree_S[100];		//観測時の全方向

	int discovery_num;		//観測回数
	int mach_z;				//対応がとれた観測の番号

	Land_Mark(){
		mach_z = -1;
		discovery_num = 0;
		for( int i=0; i<100; i++ ){
			observation_time[i] = 0;
			discovery_degree_S[i] = -1.0;
		}
	}

	Land_Mark( const Land_Mark & src ){

		myu_x = src.myu_x;
		myu_y = src.myu_y;
		myu_z = src.myu_z;
		var_11 = src.var_11;
		var_12 = src.var_12;
		var_13 = src.var_13;
		var_21 = src.var_21;
		var_22 = src.var_22;
		var_23 = src.var_23;
		var_31 = src.var_31;
		var_32 = src.var_32;
		var_33 = src.var_33;
		sec = src.sec;
		nsec = src.nsec;
		mach_z = src.mach_z;
		discovery_degree = src.discovery_degree;
		discovery_num = src.discovery_num;
		
		for( int i=0; i<128; i++ ) SIFT_Val[i] = src.SIFT_Val[i];

		for( int i=0; i<100; i++ ){
			observation_time[i]   = src.observation_time[i];
			discovery_degree_S[i] = src.discovery_degree_S[i];
		}
	}
};



//------------------------------------------------------------------------------------------------+
// 行列計算用のクラス
//------------------------------------------------------------------------------------------------+
class matrix{
 public:  
  double S00;
  double S10;
  double S01;
  double S11;
};


//------------------------------------------------------------------------------------------------+
// ランドマークと観測の対応がとれた場合のやりとり用のクラス
//------------------------------------------------------------------------------------------------+
class z_struct_3{
 public:
  double z_x;
  double z_y;
  double z_z;
  double SIFT_Val[128];
  int window_x;
  int window_y;
  bool match;
  double  match_x;
  double  match_y;
  double  match_z;
};


//------------------------------------------------------------------------------------------------+
// 行列計算用のクラス
//------------------------------------------------------------------------------------------------+
class matrix_3{
 public:  
  double S00;
  double S10;
  double S20;
  double S01;
  double S11;
  double S21;
  double S02;
  double S12;
  double S22;
};


#endif 
